import {
  pruneGmailElement,
  pruneOutlookElement,
} from "./normalizationStrategies";
import MailParser, { fixNewLines } from "../MailParser/MailParser";
import CryptoService from "../CryptoService";
import { arrayBufferToBase64 } from "../../utils/common";
import { EMAIL_VENDORS } from "../StatusesService";
import MimeVerificationService from "./index";
import { RollingHash } from "../../index";
import VerificationError from "../VerificationService/VerificationError";
import { removeSpacesAndLinebreaks } from "../../utils/stringUtils";
import { printHtmlChildren, printOutlookElement, pruneHtmlNode } from "./utils";

const vendorPruningFunctions = {
  [EMAIL_VENDORS.GMAIL]: pruneGmailElement,
  [EMAIL_VENDORS.OUTLOOK]: pruneOutlookElement,
};

const vendorPrintingFunctions = {
  [EMAIL_VENDORS.OUTLOOK]: printOutlookElement,
};

// Load JSDOM dynamically for Node env only, because built CRA is crashing with it
let JSDOM;
const loadJSDOM = async () => {
  if (!JSDOM) {
    JSDOM = (await import("jsdom")).JSDOM;
  }
};

typeof window === "undefined" && loadJSDOM();

const prepareMimeForPartsVerification = async (
  mimeString: string,
  attachmentsHashes: Array<string>,
  senderSystem: string,
  sourceHtmlRollingHash: string,
  sourceHtmlSize: number,
  sourcePlainRollingHash: string,
  sourcePlainSize: number
): Promise<string> => {
  const { htmlPart, plainPart } = getNormalizedMimeParts(
    mimeString,
    senderSystem
  );

  const originalHtml = findSubstring(
    htmlPart,
    sourceHtmlRollingHash,
    sourceHtmlSize
  );
  if (originalHtml === null) {
    throw new VerificationError(
      "Unable to verify MIME. Source HTML part was not found."
    );
  }

  const originalPlain = findSubstring(
    plainPart,
    sourcePlainRollingHash,
    sourcePlainSize
  );
  if (originalPlain === null) {
    throw new VerificationError(
      "Unable to verify MIME. Source Plain part was not found."
    );
  }

  return await MimeVerificationService.preparePartsForSigning(
    [originalHtml, originalPlain],
    attachmentsHashes
  );
};

const getNormalizedMimeParts = (
  mimeString: string,
  senderSystem: string
): { htmlPart: string; plainPart: string } => {
  const {
    htmlPart,
    plainPart,
  } = MimeVerificationService.getMimeHtmlAndPlainParts(mimeString);

  const normalizedHtmlPart = MimeVerificationService.normalizeVendorHtml(
    htmlPart,
    senderSystem
  );

  const normalizedPlainPart = removeSpacesAndLinebreaks(plainPart);

  return {
    htmlPart: normalizedHtmlPart,
    plainPart: normalizedPlainPart,
  };
};

const findSubstring = (
  string: string,
  rollingHash: string,
  substringSize: number
): string => {
  if (!string.length) {
    return "";
  }

  const startingIndex = RollingHash.searchHashInText(
    string,
    rollingHash,
    substringSize
  );

  if (startingIndex === -1) {
    return null;
  }

  return string.slice(startingIndex, startingIndex + substringSize);
};

const getMimeHtmlAndPlainParts = (
  mimeString: string
): {
  htmlPart: string;
  plainPart: string;
} => {
  mimeString = fixNewLines(mimeString);
  const mimeParts = MailParser.parseMIME(mimeString);

  // TODO: check, whether we need to support 2+ html/plain MIME parts.
  const htmlPart = MailParser.getHTML(mimeString, mimeParts);
  const plainPart = MailParser.getPlain(mimeString, mimeParts);

  return {
    htmlPart,
    plainPart,
  };
};

/**
 * @param parts - html and plain parts strings
 * @param attachmentsHashes - base64 strings of the attachments hashes
 */
const preparePartsForSigning = async (
  parts: Array<string>,
  attachmentsHashes: Array<string>
): Promise<string> => {
  let hashedParts = await Promise.all(
    parts.map(async (part) => {
      return arrayBufferToBase64(await CryptoService.SHA256(part));
    })
  );

  hashedParts = [...hashedParts, ...attachmentsHashes];

  return hashedParts.sort().join("\n");
};

const normalizeVendorHtml = (htmlString: string, vendor: string): string => {
  let document: HTMLDocument;

  if (typeof DOMParser !== "undefined") {
    const parser = new DOMParser();
    document = parser.parseFromString(htmlString, "text/html");
  } else {
    const { window } = new JSDOM(htmlString);
    document = window.document;
  }

  const mimeBody = document.firstChild.lastChild;

  const elementPruningFunction = vendorPruningFunctions[vendor];

  if (!elementPruningFunction) {
    throw new Error(
      `Vendor "${vendor}" is not supported. Please, develop a normalization strategy for it.`
    );
  }

  pruneHtmlNode(document, elementPruningFunction);

  const vendorPrintFunction = vendorPrintingFunctions[vendor];

  return printHtmlChildren(mimeBody, vendorPrintFunction, 0);
};

const normalizePlainPart = (text: string): string => {
  return removeSpacesAndLinebreaks(text);
};

export default {
  normalizeVendorHtml,
  getMimeHtmlAndPlainParts,
  preparePartsForSigning,
  prepareMimeForPartsVerification,
  getNormalizedMimeParts,
  normalizePlainPart,
};
