import { removeSpacesAndLinebreaks } from "../../utils/stringUtils";
import { escapeHtmlString } from "../..";
import { DOCUMENT_NODE, ELEMENT_NODE, TEXT_NODE } from "./constants";

export const printHtmlChildren = (
  node: Node,
  printFunction: (node: Node) => string,
  depth: number
): string => {
  let child = node.firstChild;
  if (!child) {
    return "";
  }

  if (child == node.lastChild && child.nodeType == TEXT_NODE) {
    return printHtmlNode(child, printFunction, depth);
  } else {
    let result = "";
    while (child) {
      result = result.concat(printHtmlNode(child, printFunction, depth));
      child = child.nextSibling;
    }

    return result;
  }
};

export const printHtmlNode = (
  node: Node,
  printFunction: (node: Node) => string,
  depth: number
): string => {
  let result = "";

  if (printFunction) {
    result = printFunction(node);
    if (result) {
      return result;
    }
  }

  switch (node.nodeType) {
    case TEXT_NODE:
      result += "<TEXT>";
      result += removeSpacesAndLinebreaks(node.textContent);
      result += "</TEXT>";
      result += "\n";
      break;
    case DOCUMENT_NODE:
      result += printHtmlChildren(node, printFunction, depth);
      break;
    case ELEMENT_NODE:
      result += "<" + node.nodeName;
      Array.from((node as HTMLElement).attributes)
        .sort((a, b) => a.name.localeCompare(b.name))
        .forEach((attribute) => {
          result += ` ${attribute.name}`;
          if (attribute.value) {
            result += `="${escapeHtmlString(attribute.value)}"`;
          }
        });

      if (node.firstChild) {
        result += ">";
        result += "\n";
        result += printHtmlChildren(node, printFunction, depth + 1);
        result += "</" + node.nodeName + ">";
      } else {
        result += "/>";
      }
      result += "\n";
  }

  return result;
};

export const pruneHtmlNode = (
  node: Node,
  pruneElement: (element: HTMLElement) => boolean
): boolean => {
  let toBeRemoved = false;

  switch (node.nodeType) {
    case node.COMMENT_NODE:
    case node.DOCUMENT_TYPE_NODE:
      toBeRemoved = true;
      break;
    case node.TEXT_NODE: {
      const trimmedText = node.textContent.trim();
      if (trimmedText === "") {
        toBeRemoved = true;
      } else {
        node.textContent = trimmedText;
      }
      break;
    }
    case node.ELEMENT_NODE:
      toBeRemoved = pruneElement(node as HTMLElement);
  }

  if (toBeRemoved) {
    return true;
  }

  const childrenToRemove = [];
  let child = node.firstChild;

  while (child) {
    pruneHtmlNode(child, pruneElement) && childrenToRemove.push(child);
    child = child.nextSibling;
  }

  childrenToRemove.forEach((child) => node.removeChild(child));

  return false;
};

export const printOutlookElement = (node: Node): string => {
  if (node.nodeType === ELEMENT_NODE) {
    if ((node as HTMLElement).classList.contains("WordSection1")) {
      return printHtmlChildren(node, null, 0);
    }
  }
};
